/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.parsing.binding;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.pvm.internal.util.XmlUtil;
import org.ow2.bonita.pvm.internal.xml.Parse;
import org.ow2.bonita.pvm.internal.xml.Parser;
import org.ow2.bonita.facade.def.element.FormalParameterDefinition;
import org.ow2.bonita.facade.def.element.TimeEstimationDefinition;
import org.ow2.bonita.facade.def.element.impl.TimeEstimationDefinitionImpl;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivitySetDefinition;
import org.ow2.bonita.facade.def.majorElement.ApplicationDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.PackageDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.TransitionDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessFullDefinitionImpl;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.parsing.binding.PackageBinding.PackageDescription;
import org.ow2.bonita.services.util.ServiceEnvTool;
import org.w3c.dom.Element;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class WorkflowProcessBinding extends MajorElementBinding {

  private static final Logger LOG = Logger.getLogger(PackageBinding.class.getName());

  public WorkflowProcessBinding() {
    super("WorkflowProcess");
  }

  public Object parse(final Element workflowProcessElement, final Parse parse, final Parser parser) {
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("parsing element = " + workflowProcessElement);
    }

    final PackageDefinitionUUID packageUUID = getObject(PackageDefinitionUUID.class, parse);
    @SuppressWarnings("unused")
    final
    String packageName = getObject(PackageDescription.class, parse).getPackageName();

    final String id = getId(workflowProcessElement);
    final String name = getName(workflowProcessElement);


    final Element processHeaderElement = XmlUtil.element(workflowProcessElement, "ProcessHeader");
    final String created = getChildTextContent(processHeaderElement, "Created");
    final String description = getChildTextContent(processHeaderElement, "Description");
    final String priority = getChildTextContent(processHeaderElement, "Priority");
    final String limit = getChildTextContent(processHeaderElement, "Limit");
    final String validFrom = getChildTextContent(processHeaderElement, "ValidFrom");
    final String validTo = getChildTextContent(processHeaderElement, "ValidTo");
    final ProcessDefinition.DurationUnit durationUnit = getEnumValue(
        ProcessDefinition.DurationUnit.class, XmlUtil.attribute(processHeaderElement, "DurationUnit"), null);

    TimeEstimationDefinition timeEstimation = null;
    if (processHeaderElement != null) {
      final Element timeEstimationElement = XmlUtil.element(processHeaderElement, "TimeEstimation");
      if (timeEstimationElement != null) {
        final String duration = getChildTextContent(timeEstimationElement, "Duration");
        final String waitingTime = getChildTextContent(timeEstimationElement, "WaitingTime");
        final String workingTime = getChildTextContent(timeEstimationElement, "WorkingTime");
        timeEstimation = new TimeEstimationDefinitionImpl(duration, waitingTime, workingTime);
      }
    }

    final Element redefinableHeaderElement = XmlUtil.element(workflowProcessElement, "RedefinableHeader");
    final String author = getChildTextContent(redefinableHeaderElement, "Author");
    String version = getChildTextContent(redefinableHeaderElement, "Version");
    if (version == null || "".equals(version)) {
      //get package version
      version = parse.findObject(PackageDescription.class).getVersion();
    }
    final String codepage = getChildTextContent(redefinableHeaderElement, "Codepage");
    final String countrykey = getChildTextContent(redefinableHeaderElement, "Countrykey");
    PackageDefinition.PublicationStatus publicationStatus = null;
    if (redefinableHeaderElement != null) {
      publicationStatus = getEnumValue(PackageDefinition.PublicationStatus.class,
          XmlUtil.attribute(redefinableHeaderElement, "PublicationStatus"), null);
    }

    final ProcessDefinitionUUID recordUUID =
      ServiceEnvTool.getUUIDGenerator().getProcessDefinitionUUID(packageUUID, id, version);
    parse.pushObject(recordUUID);

    Set<String> responsibles = null;
    if (redefinableHeaderElement != null) {
      final Element responsiblesElement = XmlUtil.element(redefinableHeaderElement, "Responsibles");
      if (responsiblesElement != null) {
        final List<Element> responsibleElements = XmlUtil.elements(responsiblesElement, "Responsible");
        responsibles = new HashSet<String>();
        for (final Element responsibleElement : responsibleElements) {
          responsibles.add(responsibleElement.getTextContent());
        }
      }
    }

    final ProcessDefinition.AccessLevel accessLevel = getEnumValue(
        ProcessDefinition.AccessLevel.class, XmlUtil.attribute(workflowProcessElement, "AccessLevel"), null);
    final List<FormalParameterDefinition> formalParameters = getFormalParameters(workflowProcessElement, parse, parser);

    final Set<ActivitySetDefinition> activitySets = parseXpdlMajorElementList(workflowProcessElement,
        "ActivitySets", "ActivitySet", parse, parser);
    final Set<ParticipantDefinition> participants = parseXpdlMajorElementList(workflowProcessElement,
        "Participants", "Participant", parse, parser);
    final Set<ApplicationDefinition> applications = parseXpdlMajorElementList(workflowProcessElement,
        "Applications", "Application", parse, parser);
    final Set<DataFieldDefinition> allDataFields = parseXpdlMajorElementList(workflowProcessElement,
        "DataFields", "DataField", parse, parser);

    final Set<DataFieldDefinition> processDataFields = new HashSet<DataFieldDefinition>();
    final Set<DataFieldDefinition> activityDataFields = new HashSet<DataFieldDefinition>();
    if (allDataFields != null) {
      for (final DataFieldDefinition dataField : allDataFields) {
        if (dataField.isActivityDataField()) {
          activityDataFields.add(dataField);
        } else {
          processDataFields.add(dataField);
        }
      }
    }
    Set<ActivityDefinition> activities = null;
    try {
      parse.pushObject(activityDataFields);
      activities = parseXpdlMajorElementList(workflowProcessElement,"Activities", "Activity", parse, parser);
    } finally {
      parse.popObject();
    }
    final Set<TransitionDefinition> transitions = parseXpdlMajorElementList(workflowProcessElement,
        "Transitions", "Transition", parse, parser);



    final ProcessDefinition processDefinition = new ProcessFullDefinitionImpl(recordUUID, packageUUID,
        accessLevel, author, codepage, countrykey, created, description, durationUnit,
        formalParameters, id, limit, name, priority, publicationStatus, responsibles, timeEstimation,
        validFrom, validTo, version,
        processDataFields, participants,
        activities, transitions,
        activitySets, applications, ProcessState.DEPLOYED);

    parse.popObject();
    return processDefinition;
  }

}


